{*****************************************************************************
*                                                                            *
*  TEXMAP.DPR                                                                *
*  TEXMAPU.PAS                                                               *
*                                                                            *
*  This program uses Fastgraph's 3D geometry system to draw a texture-mapped *
*  cube. Uses the approach of projecting the entire cube in one operation to *
*  take advantage of shared vertices between faces.                          *
*                                                                            *
*****************************************************************************}

unit TexmapU;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, FGWin;

type
  TForm1 = class(TForm)
    procedure AppOnActivate(Sender: TObject);
    procedure AppIdle(Sender: TObject; var Done: Boolean);
    procedure FormActivate(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}
type
  FACE = array [1..4] of integer;
  FIXED = longint;
  POINT2D = record
    x : integer;
    y : integer;
  end;
  POINT3D = record
    x : FIXED;
    y : FIXED;
    z : FIXED;
  end;

const
  { virtual buffer dimensions }
  vbWidth  : integer = 640;
  vbHeight : integer = 480;

  { vertices of a 40x40x40 cube, defined in 3D coordinates }
  cube : array [0..7] of POINT3D = (
    (x:$FFEC0000; y:$00140000; z:$00140000),
    (x:$00140000; y:$00140000; z:$00140000),
    (x:$FFEC0000; y:$FFEC0000; z:$00140000),
    (x:$00140000; y:$FFEC0000; z:$00140000),
    (x:$FFEC0000; y:$00140000; z:$FFEC0000),
    (x:$00140000; y:$00140000; z:$FFEC0000),
    (x:$FFEC0000; y:$FFEC0000; z:$FFEC0000),
    (x:$00140000; y:$FFEC0000; z:$FFEC0000));

  { indices into the cube array that define each face of the cube }
  face1 : FACE = (0,1,3,2);
  face2 : FACE = (1,5,7,3);
  face3 : FACE = (6,7,5,4);
  face4 : FACE = (4,0,2,6);
  face5 : FACE = (4,5,1,0);
  face6 : FACE = (2,3,7,6);

  { for convenience, an array of pointers to each of the six faces }
  faces : array [1..6] of ^FACE = (@face1,@face2,@face3,@face4,@face5,@face6);

  { width of texture map in pixels }
  TMAP_WIDTH = 128;

  { coordinates defining source polygon vertices within the texture map array }
  source : array [1..8] of integer = (
    0,0, TMAP_WIDTH-1,0, TMAP_WIDTH-1,TMAP_WIDTH-1, 0,TMAP_WIDTH-1);

var
  { 3D transformation matrix }
  transform : array [1..12] of FIXED;

  { texture map array }
  texture : array [1..TMAP_WIDTH*TMAP_WIDTH] of byte;

  cxClient, cyClient : integer;
  dc   : hDC;
  hpal : hPalette;
  hvb  : integer;
  redo : boolean;
  xAngle, yAngle, zAngle : integer;

{*****************************************************************************
*                                                                            *
*  draw_cube()                                                               *
*                                                                            *
*  Draws each of the six cube faces. For each face, we obtain the projected  *
*  coordinates from the projected_cube array and store them in the vertices  *
*  array so we can draw the cube face with fg_texmap().                      *
*                                                                            *
*****************************************************************************}

procedure draw_cube;

var
  i, j : integer;
  vertices : array [1..4] of POINT2D;
  projected_cube : array [0..7] of POINT2D;

begin
  fg_project(transform,cube,projected_cube,8);
  for i := 1 to 6 do
  begin
    for j := 1 to 4 do
      vertices[j] := projected_cube[faces[i]^[j]];
    fg_texmap(vertices,source,4);
  end;
end;

{*****************************************************************************
*                                                                            *
*  check_for_movement()                                                      *
*                                                                            *
*  The check_for_movement() function checks if any of the four arrow keys    *
*  are pressed, and if so, pans in that direction if possible. It is called  *
*  from the application's OnIdle event handler.                              *
*                                                                            *
*****************************************************************************}

procedure check_for_movement;

var
  shift_key : boolean;

begin
  { erase the previous frame from the virtual buffer }
  fg_setcolor(255);
  fg_fillpage;

  { project the cube onto the virtual buffer and draw it }
  draw_cube;

  { display what we just drew }
  fg_vbscale(0,vbWidth-1,0,vbHeight-1,0,cxClient-1,0,cyClient-1);

  { check if either shift key is pressed }
  shift_key := (fg_kbtest(42) = 1) or (fg_kbtest(54) = 1);

  { + and - move cube along the z axis (+ is toward viewer, - is }
  { away from viewer), restricting movement beyond z=-40 }
  if (fg_kbtest(74) = 1) then
    fg_move3d(transform,0,0,$FFFD0000,1)
  else if (fg_kbtest(78) = 1) then
  begin
    if (transform[12] < $FFD80000) then
      fg_move3d(transform,0,0,$00030000,1);
  end

  { left and right arrow keys move cube along x axis }
  else if (fg_kbtest(75) = 1) then
    fg_move3d(transform,$FFFD0000,0,0,1)
  else if (fg_kbtest(77) = 1) then
    fg_move3d(transform,$00030000,0,0,1)

  { up and down arrow keys move cube along y axis }
  else if (fg_kbtest(72) = 1) then
    fg_move3d(transform,0,$00030000,0,1)
  else if (fg_kbtest(80) = 1) then
    fg_move3d(transform,0,$FFFD0000,0,1)

  { x rotates counterclockwise around x axis, X rotates clockwise }
  else if (fg_kbtest(45) = 1) then
  begin
    if (shift_key) then
    begin
      inc(xAngle,6);
      if (xAngle >= 360) then dec(xAngle,360);
    end
    else
    begin
      dec(xAngle,6);
      if (xAngle < 0) then inc(xAngle,360);
    end;
    redo := True;
  end

  { y rotates counterclockwise around y axis, Y rotates clockwise }
  else if (fg_kbtest(21) = 1) then
  begin
    if (shift_key) then
    begin
      inc(yAngle,6);
      if (yAngle >= 360) then dec(yAngle,360);
    end
    else
    begin
      dec(yAngle,6);
      if (yAngle < 0) then inc(yAngle,360);
    end;
    redo := True;
  end

  { z rotates counterclockwise around z axis, Z rotates clockwise }
  else if (fg_kbtest(44) = 1) then
  begin
    if (shift_key) then
    begin
      inc(zAngle,6);
      if (zAngle >= 360) then dec(zAngle,360);
    end
    else
    begin
      dec(zAngle,6);
      if (zAngle < 0) then inc(zAngle,360);
    end;
    redo := True;
  end;

  { if rotation changed, modify the transformation matrix accordingly }
  if (redo) then
  begin
    fg_rotate3d(transform,xAngle*10,yAngle*10,zAngle*10,0);
    redo := False;
  end;

end;

{****************************************************************************}

procedure TForm1.AppOnActivate(Sender: TObject);
begin
  fg_realize(hpal);
  Invalidate;
end;

procedure TForm1.AppIdle(Sender: Tobject; var Done: Boolean);
begin
  check_for_movement;
  Done := False;
end;

procedure TForm1.FormActivate(Sender: TObject);
begin
  fg_realize(hpal);
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  dc := GetDC(Form1.Handle);
  fg_setdc(dc);
  hpal := fg_defpal;
  fg_realize(hpal);

  fg_vbinit;
  hvb := fg_vballoc(vbWidth,vbHeight);
  fg_vbopen(hvb);
  fg_vbcolors;

  { define 3D viewport and initial transformation matrix }
  fg_view3d(0,vbWidth-1,0,vbHeight-1,$00008000);
  fg_rotate3d(transform,0,0,0,0);
  fg_move3d(transform,0,0,$FF9C0000,0);

  { obtain the texture map from the MOUSE.PCX file }
  fg_showpcx('MOUSE.PCX'+chr(0),0);
  fg_move(80,150);
  fg_getimage(texture,TMAP_WIDTH,TMAP_WIDTH);
  fg_invert(texture,TMAP_WIDTH,TMAP_WIDTH);
  fg_texture(texture,TMAP_WIDTH);

  fg_setcolor(255);
  fg_fillpage;

  redo := False;
  xAngle := 0;
  yAngle := 0;
  zAngle := 0;

  Application.OnActivate := AppOnActivate;
  Application.OnIdle := AppIdle;
end;

procedure TForm1.FormPaint(Sender: TObject);
begin
  fg_vbscale(0,vbWidth-1,0,vbHeight-1,0,cxClient-1,0,cyClient-1);
end;

procedure TForm1.FormResize(Sender: TObject);
begin
  cxClient := ClientWidth;
  cyClient := ClientHeight;
  Invalidate;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  fg_vbclose;
  fg_vbfree(hvb);
  fg_vbfin;
  DeleteObject(hpal);
  ReleaseDC(Form1.Handle,dc);
end;

end.
